//
// Debug and bug identification helpers
//
// Author: Jonas Tingeborn
//
openinchrome.Debug = (function(){
  "use strict"; 
  var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService),
      isTraceEnabled = false;

  function alert(msg){
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    promptService.alert(window, "", msg);
  }
  function log(){
    if(Firebug.Console) Firebug.Console.log.apply(Firebug.Console,arguments);
  }
  
  // Flattens prototype hierarchy and removes cyclic references from self 
  // and nested objects
  function flatten(o,refed){
    var h={}, p, v;
    refed = refed || [];                       // If called without second array parameter, create an empty array instead, to allow proper recursion with breaking of cyclic references.
    refed.push(o);                             // Mark the as processed, so we can detect indications of cyclic refs.
    while(o) {                                 // Process each object in the prototype chain, from bottom most descendant to top most ancestor.
      for(var p in o){                         // Go through each property on the current (context) object.
        if( /^[A-Z0-9_]+$/.test(p) ) continue; // Filter out noise by omitting to include constants.
        try {                                  // Some properties are protected by mozilla, so just skip them.
          if(!h[p]) {                          // Avoid overwriting a descendent's property value with that of an ancestor (prototype object).
            v = o[p];                          // The property's value (short hand).
            if(v instanceof Window || v instanceof Node || typeof(v) == "function") continue;
            if(typeof(v) == "object") {        // Special handling (care) if the property value is an object.
              if(refed.indexOf(v)<0) {         // Avoid creating cyclic object references.
                refed.push(v);                 // Mark the as processed, so we can detect indications of cyclic refs.
                v = flatten(v,refed);          // Recurse into the property value object and flatten that as well.
                h[p] = v                       // Assign the flattened and non-cyclic flattened value, in place of the original value.
              }
            } else {
              h[p] = v                         // The property value is of a type that is OK to serialize to JSON, so just copy it.
            }
          }
        } catch(e) {}
      }
      o = o.__proto__;                         // Walk the prototype chain
    }
    return h;                                  // Return a variant of "o", but that is serializable in its entirety (entire depth / object graph).
  }

  // A JSON serializer function that also works with native browser objects
  // (by discarding anything that might cause the JSON serializer to barf)
  function convertToJSON(o) {
    return JSON.stringify( flatten(o) );       // Make sure the object is serializable before letting the JSON serializer at it.
  }
  
  function trace(){
    if(!isTraceEnabled) return;
    var args = [];
    for(var i=0; i<arguments.length; i++){
      if( "object" == typeof(arguments[i]) ) args.push( convertToJSON(arguments[i]) );
      else args.push( arguments[i].toString() );
    }
    consoleService.logStringMessage("openinchrome: " + args.join(" "));
  }

  return {
    trace: trace,
    alert: alert,
    log: log,
    setTracing: function(enabled) {
      isTraceEnabled = enabled;
    }
  }
}());
