//
// The value added functionality of this addon (the domain logic)
//
// Author: Jonas Tingeborn
//
var openinchrome = (function(){
  "use strict";
  var opts, tabListener, trace;   // Are assigned onLoad
  const INVALID_CONTEXT_TAGS = ["a","img","input","textarea","button"];

  ////
  // URL Filter, to ensure only applicable URLs are sent to chrome
  function isProtocolSupported(scheme){
    return /^(https?|file|ftp):?$/.test(scheme);
  }

  function isUrlSupported(url){
    if(!url) return false;
    var proto = url.split(":")[0];
    return isProtocolSupported(proto);
  }
  function log(){
    if(typeof Firebug != "undefined" && Firebug.Console) Firebug.Console.log.apply(Firebug.Console,arguments);
    //if(Firebug.Console) { var args = []; for(var i=0;i<arguments.length;i++) args.push(arguments[i]);  Firebug.Console.log(args); }
  }
  function alert(msg){
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    promptService.alert(window, "", msg);
  }
  // Returns the closest hierarchy element of type tagName or null if el
  // isn't a or isn't contained in a tagName
  function getClosest(el, tagName) {
    while(el) {
      if(el.localName && el.localName.toLowerCase() == tagName) return el;
      el = el.parentNode;
    }
  }
  function isOrHasParent(el, tagName) {
    return !!getClosest(el, tagName);
  }

  ////
  // DOM Event helper
  function isApplicableClickEvent(e, opts){
    var button;
    if     (e.button == 0 ) button = "left";
    else if(e.button == 1 ) button = "middle";
    else if(e.button == 2 ) button = "right";
    else button = "";

    return opts.button == button    &&
           opts.ctrl   == e.ctrlKey &&
           opts.alt    == e.altKey  &&
           opts.shift  == e.shiftKey;
  };

  ////
  // Window
  function getMainWindow(){
    return window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                           .getInterface(Components.interfaces.nsIWebNavigation)
                           .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
                           .rootTreeItem
                           .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                           .getInterface(Components.interfaces.nsIDOMWindow);
  }

  ////
  // Tab helpers
  function getTabs(){
    var ret = [], tabs = gBrowser.tabContainer, i, tab;
    for(i=0; i < tabs.itemCount; i++){
      tab = tabs.getItemAtIndex( i );
      ret.push(tab);
    }
    return ret;
  }
  function getUrlFromTab(tab) {
    return gBrowser.getBrowserForTab(tab).contentDocument.location.href;
  }
  function getActiveTab(){
    return gBrowser.selectedTab;
  }
  function isCurrentTabApplicable(){
    try {
      return isProtocolSupported(content.location.protocol);  // Only protocols supported by chrome.
    } catch (ignore) {}                                       // Content window not yet in a state to report a loaded page, so the addon isn't yet applicable for invocation.
    return false;
  }

  ////
  // Addon icon helpers
  function enableIcon(enabled){
    try {
      document.getElementById('openinchrome-btn').disabled = !enabled;
    } catch(ignore) { /* Not yet properly initialized */ }
  }

  ////
  // Menu functions
  function showUIElement(id, visible){
    getMainWindow().document.getElementById(id).className = visible ? "" : "hide";
  }
  function showToolbar(visible){
    showUIElement("menu_openinchrome", visible);
  }
  function showContextMenu(visible){
    showUIElement("ctxmenu_openinchrome", visible);
  }
  function showLinkContextMenu(visible){
    showUIElement("ctxmenu_openinchrome_link", visible);
  }

  ////
  // Addon UI widget state management (visibility / enabled or disabled)
  function updateCommandWidgets(){
    var applicable = isCurrentTabApplicable();
    showToolbar    ( applicable && opts.isToolbarEnabled()     );
    showContextMenu( applicable && opts.isContextMenuEnabled() );
    enableIcon( applicable );
  }

  ////
  // Native Process
  function openUrlInChrome(url){
    if(! opts.ensureConfigured() ) return;
    trace("7: Addon is configured");

    var args    = opts.getPathAsCommandArray().concat(url),
        chrome  = args.shift(),
        file    = window.Components.classes['@mozilla.org/file/local;1'].createInstance(Components.interfaces.nsILocalFile),
        process = window.Components.classes['@mozilla.org/process/util;1'].createInstance(Components.interfaces.nsIProcess);

    trace("8: Launching Chrome, args = ", {chrome:chrome, url:url});
    file.initWithPath(chrome);
    process.init(file);
    process.run(false, args, args.length);
    trace("9: After invoking Chrome");
  }

  ////
  // Command handlers
  function migrateTab(tab, variant){
    var url = getUrlFromTab(tab), supported = isUrlSupported(url);
    trace("5"+ variant +": isUrlSupported ["+ url +"] = "+ supported);
    if(supported) {                                                    // Ensure only URLs with valid protocols get forwarded to Chrome
      if(tab.pinned == opts.isIgnorePinnedEnabled()) {
        trace("6"+ variant +": ignoring migrating pinned tab ["+ url +"]");
      } else {
        trace("6"+ variant +": openUrlInChrome ["+ url +"]");
        openUrlInChrome(url);
        return true;
      }
    }
  }
  function closeTab(tab, variant, closeMethod) {
    var closeOnMigrate = opts[closeMethod](), url = getUrlFromTab(tab),
        ignorePinned = opts.isIgnorePinnedEnabled();
    trace("10"+ variant +": "+ closeMethod +" = "+ closeOnMigrate +", isLastTab = "+ (getTabs().length <= 1) +
          ", ignorePinned="+ ignorePinned);
    if( tab.pinned && ignorePinned ) {
      trace("11a tab is pinned, so skipping closing it");
    } else if ( closeOnMigrate && getTabs().length > 1) {
      trace("11"+ variant +": closing tab ["+ url +"]");
      gBrowser.removeTab(tab);
      trace("12"+ variant +": tab closed ["+ url +"]");
    }
  }
  function openCurrentTabCommand(e) {
    var tab = getActiveTab();
    trace("1a: migrate current tab invoked from menu");
    if( migrateTab(tab,"a") ) {
      closeTab(tab, "a","isCloseSingleEnabled");
    }
  }
  function openAllTabsCommand(e) {
    var toClose = [];
    trace("1b: migrate all tabs invoked from menu");
    getTabs().forEach(function(tab){
      if( migrateTab(tab,"b") ) toClose.push(tab);
    });
    toClose.reverse().forEach(function(tab){       // Need to close in tab reverse order, or FF throws an exception (probably some tab index mismatch happening in the internal FF implementation)
      closeTab(tab, "b","isCloseAllEnabled");
    });
  }
  function openLinkCommand(e) {
    var url, a = getClosest(document.popupNode,"a");
    if(!a) return;
    url = a.href;
    trace("1c: Open link from context menu triggered [url="+ url +"]");
    if(isUrlSupported(url)) {
      trace("2c: URL is supported");
      openUrlInChrome(url);
    } else {
      trace("2c: Not opening link as URL is not supported [url="+ url +"]");
    }
  }

  ////
  // Tab Event handlers
  // Constructor function for MDN boiler plate when listening on tab location changes
  function newTabLocationListener(onLocationChangeCallback){
    return {
      onLocationChange: function(aProgress, aRequest, aURI) {
        onLocationChangeCallback(aProgress, aRequest, aURI);
      },
      // Interface boiler plate, doesn't seem needed in FF 11+, but retaining for now as per MDN instruction...
      QueryInterface: function(aIID) {
       if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
           aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
           aIID.equals(Components.interfaces.nsISupports))
         return this;
       throw Components.results.NS_NOINTERFACE;
      },
      onStateChange: function(aWebProgress, aRequest, aFlag, aStatus) {},
      onProgressChange: function(aWebProgress, aRequest, curSelf, maxSelf, curTot, maxTot) { },
      onStatusChange: function(aWebProgress, aRequest, aStatus, aMessage) { },
      onSecurityChange: function(aWebProgress, aRequest, aState) { }
    };
  }

  ////
  // Web page functions (content window)
  function registerPageListener(event) {
    gBrowser.addEventListener('click', handlePageClick, true);
  }
  function unregisterPageListener() {
    gBrowser.removeEventListener('click', handlePageClick, true);
  }

  // Click handler for the user's web pages with action to take evaluated from the nature of the click.
  function handlePageClick(event) {
    var doc = event.originalTarget.ownerDocument, lOpts, link;
    if (doc && doc instanceof HTMLDocument) {
      trace("==[ Web page clicked ] ================");
      trace("1: Event = ", event);
      if( opts.isMigrateLinksEnabled() ){
        lOpts = opts.getClickConfigForCommand("openlinks");
        trace("2: Migrate links option is enabled in preferences, link options = ", lOpts, "event: ", event);
        if( isApplicableClickEvent(event, lOpts) ) {
          trace("3: Click event is applicable");
          event.stopPropagation();                  // Prevent default event handler from being invoked
          link = getClosest(event.originalTarget, "a");
          if(link) {
            trace("4: Click was within a link element");
            if (link.href && link.href.length > 0) {
              trace("5: Link has a URL");
              if( isUrlSupported(link.href) ) {
                trace("6: Opening link in Chrome: ", link.href);
                openUrlInChrome(link.href);
              } else {
                trace("6F: Not acting on link, unsupported protocol: ", link.href);
              }
            } else {
              trace("5F: Not acting on link, as it has no URL");
            }
          } else {
            trace("4F: Not handling this event, click was not within a link element");
          }
        } else {
          trace("3F: Not handling this event, user pref config does not match for event");
        }
      } else {
        trace("2F: Not handling the click event, migrate links option is disabled in preferences");
      }
    }
  }
  ////////////////////

  // Opens the trace window if user has set the "debug" preference to true
  function openTraceWindow() {
    if(opts.isDebugEnabled()) {
      window.open("chrome://openinchrome/content/tracewindow.xul", "openinchrome-tracewindow",
                  "chrome,titlebar,toolbar,centerscreen,resizable=yes,dependent=yes");
    }
  }

  ////
  // Public methods
  return {
    onLoad: function(event) {
      var onStateUpdate = function(){
        updateCommandWidgets();                                    // Update menu state whenever user changes prefs
        openTraceWindow();                                         // Open trace window if user has set that pref.
        openinchrome.Debug.setTracing(opts.isDebugEnabled());      // Do tracing only when necessary, to save CPU cycles.
      };
      tabListener = newTabLocationListener(updateCommandWidgets);
      trace = openinchrome.Debug.trace;
      opts  = openinchrome.Opts;

      opts.startPrefsMonitoring(onStateUpdate);
      gBrowser.addProgressListener(tabListener);                   // Update menu state whenever user or browser changes tab location
      onStateUpdate();
      window.addEventListener("popupshowing",openinchrome.onContextMenuPopup,false);
      registerPageListener(event);
    },
    onUnload: function() {
      opts.stopPrefsMonitoring();                                  // Stop monitoring prefs changes
      gBrowser.removeProgressListener(tabListener);                // Stop monitoring tab location changes
      unregisterPageListener();                                    // Stop monitoring clicks on web pages
    },

    ////
    // Public command handlers
    openAllTabsCommand: openAllTabsCommand,
    openCurrentTabCommand: openCurrentTabCommand,
    openLinkCommand: openLinkCommand,

    // Ensure only valid contexts trigger the addon to show its context menu item
    onContextMenuPopup: function(event) {
      var applicable = isCurrentTabApplicable(),
          el = document.popupNode, i;

      if(isOrHasParent(el, "a")) {                         // handle links explicitly as per feature request
        showLinkContextMenu( true );
      } else {
        showLinkContextMenu( false );
      }

      for(i=0; i < INVALID_CONTEXT_TAGS.length; i++){      // For all other contexts, allow showing menu unless context is black listed
        if( isOrHasParent(el, INVALID_CONTEXT_TAGS[i]) ) {
          showContextMenu( false );
          return;
        }
      }
      if(getBrowserSelection().length > 0) {               // If valid context, but user has selected text, it's unlikely the user wants to see the menu
        showContextMenu( false );
      } else {
        showContextMenu( applicable && opts.isContextMenuEnabled() );
      }
    },

    ////
    // Button click handler, action depends on how the click was performed.
    buttonClickCommand: function(e) {
      trace("==[ Button clicked ] ================");
      trace("1: isCurrentTabApplicable, Event = ", e);

      if( !isCurrentTabApplicable() ) return;                      // Explicitly No-op the command button behavior, since FF won't do it for us, even if the icon is disabled.

      trace("2: After isCurrentTabApplicable");

      var ctCfg = opts.getClickConfigForCommand("opencurrenttab"),
          atCfg = opts.getClickConfigForCommand("openalltabs")

      trace("3: isApplicable event check [opencurrenttab key config] = ", ctCfg, " | [openalltabs key config] = ", atCfg);
      if ( isApplicableClickEvent(e, ctCfg) ) {
        trace("4a: openCurrentTabCommand");
        openCurrentTabCommand(e);
      } else if ( isApplicableClickEvent(e, atCfg) ) {
        trace("4b: openAllTabsCommand");
        openAllTabsCommand(e);
      }
    }

  };
})();
window.addEventListener("load", openinchrome.onLoad, false);
window.addEventListener("unload", openinchrome.onUnload, false);